c========================================================================        
c                    **** utilio.f  ****
c========================================================================        
c
c  utilio.f contains routines and functions for I/O and related stuff.
c  Currently contains the routines:
c      unpack  : reads in and unpacks control parameters
c      strings : sets up character strings that name the method 
c                and preconditioner used; for Xgraph output files
c
c      matchk  : checks matrix and inserts diagonal entries if needed.
c      readf   : read in a vector stored in ascii, one entry per row
c      readmt3 : version of readmt from SPARSKIT which uses pointers into
c                a work array, handles rectangular matrices, and provides
c                space for diagonal entry in each row if specified.
c      rdline  : reads in a single line of input for readmt3
c      rlunf   : Read a "line" from an unformatted file.
c      readunf2: Read a H/B matrix stored in an unformatted file.
c      prtval  : Print out a vector (int or dbl)
c      prtmt   : Print out a Harwell/Boeing matrix (this is from Sparskit)
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     Phone: (812) 855-7790
c     email: bramley@cs.indiana.edu
c     Mon Dec 16 11:33:48 EST 1996
c
c
c========================================================================        

c**********************************************************************
c                   SPLIB Copyright (C) 1995                         **
c                     Indiana University                             **
c**********************************************************************

      subroutine strings(solmeth, solver, pcmeth, precond)
c-----------------------------------------------------------------------
c  Set up character strings identifying the solver/precond pair as
c  needed in SPLIB's Xgraph output files.
c-----------------------------------------------------------------------
      implicit none
      character solver*5, precond*4
      integer solmeth, pcmeth

      if( solmeth .eq. 1) then
         solver = 'BiCG '
      elseif (solmeth .eq. 2) then
         solver = 'CGNE '
      elseif (solmeth .eq. 3) then
         solver = 'CGNR '
      elseif (solmeth .eq. 4) then
         solver = 'CG-Sq'
      elseif (solmeth .eq. 5) then
         solver = 'CGStb'
      elseif (solmeth .eq. 6) then
         solver = 'GMRES'
      elseif (solmeth .eq. 7) then
         solver = 'QMRTF'
      elseif (solmeth .eq. 8) then
         solver = 'BiCGS'
      elseif (solmeth .eq. 9) then
         solver = 'GMRST'
      elseif (solmeth .eq. 10) then
         solver = 'JACOB'
      elseif (solmeth .eq. 11) then
         solver = 'GAUSS'
      elseif (solmeth .eq. 12) then
         solver = 'SOR'
      elseif (solmeth .eq. 13) then
         solver = 'ORTM'
      elseif (solmeth .eq. 14) then
         solver = 'CSCGS'
      end if
c
      if (pcmeth .eq. 1) then
        precond = 'ILU '
      elseif (pcmeth .eq. 2) then
        precond = 'MILU'
      elseif (pcmeth .eq. 3) then
        precond = 'ILUT'
      elseif (pcmeth .eq. 4) then
        precond = 'SSOR'
      elseif (pcmeth .eq. 5) then
        precond = 'TRID'
      elseif (pcmeth .eq. 6) then
        precond = 'ILU0'
      elseif (pcmeth .eq. 7) then
        precond = 'EMGS'
      elseif (pcmeth .eq. 0) then
        precond = 'NOPC'
      end if
c
	return
	end
c========================================================================

       subroutine matchk(a, ja, ia, n, diag, ndiag)
c------------------------------------------------------------------------
c
c  Check the matrix A in CSR format to make sure it includes all diagonal
c  entries.  If it does not have them, insert 0.0 into the data structure
c  for each missing diagonal entry.  At the same time, sort the rows of
c  A to have entries in increasing column order.  This is useful for 
c  preconditioning methods like SSOR or ILU(-1), and for matrix splitting
c  based iterative methods like Jacobi.
c
c  a, ja, ia	: matrix A in CSR format
c  n		: order of matrix A
c  diag		: integer work vector of length n.  On return, diag(1:n)
c                 contains pointers to the index of each diagonal entry
c                 in the modified CSR data structure.
c  ndiag	: integer, on return gives number of diag entries that
c                 were missing from original matrix data structure.
c  
c------------------------------------------------------------------------
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Thu Aug 25 08:06:44 EST 1994
c
c------------------------------------------------------------------------
       implicit none

c      ------------------------
c      Declaration of arguments
c      ------------------------
       real*8  a(*)
       integer n, ja(*), ia(n+1), diag(n), ndiag

c      ---------------
c      Local variables
c      ---------------
       integer row, lastptr, ibegin, iend, j, lp, ap, added

c      --------------------------------------------
c      Value to insert for missing diagonal entries
c      --------------------------------------------
       real*8 INSERT 
c       parameter (INSERT = 1.0d0)
       parameter (INSERT = 0.0d0)

c=======================================================================
c          Beginning of Executable Statements
c=======================================================================


c      ----------------------------------------------------------------
c      Find diagonal entry for each row.  If it exists, set diag(row)
c      to be its index in CSR data structure.  Otherwise, set diag(row)
c      to be zero.
c      ----------------------------------------------------------------
       ndiag = 0
       do row = 1, n
          diag(row) = 0
          do j = ia(row), ia(row+1)-1
             if (ja(j) .eq. row) then
                diag(row) = j
c          a(j) = max(a(j),1.0d0)
                go to 20
             end if 
          end do
          ndiag = ndiag + 1
 20    continue
       end do

c      --------------------------------------------------------
c      If ndiag = 0 here, then all diagonal entries are present
c      in the data structure.  Sort the rows and exit.
c      --------------------------------------------------------
       if (ndiag .eq. 0) then
          do row = 1, n
             ibegin = ia(row)
             iend   = ia(row+1)
             call rowinc( iend - ibegin, ja(ibegin), a(ibegin) )
          end do
          return
       end if

c      ---------------------------------------------------------
c      Otherwise, insert some diagonal entries into data struct.
c      ---------------------------------------------------------
       lastptr = ia(n+1) - 1 + ndiag
       lp = lastptr
       ap = ia(n+1) - 1

       do row = n, 1, -1

          ibegin = ia(row)
          iend   = ia(row+1)

c         --------------------------------------------
c         Copy a, ja over, shifting to right as needed
c         --------------------------------------------
          do j = iend-1, ibegin, -1
             a(lp)  = a(ap)
             ja(lp) = ja(ap)
             lp = lp - 1 
             ap = ap - 1 
          end do 

c         --------------------------------
c         Insert diagonal entry if missing
c         --------------------------------
          added = 0
          if (diag(row) .eq. 0) then
             a(lp) = INSERT
             ja(lp) = row
             lp  = lp - 1
             added = 1
          end if 

c         ----------------------------
c         Sort newly modified row of A
c         ----------------------------
          call rowinc( iend - ibegin + added, ja(lp+1), a(lp+1) )
c  check for duplicates
c                if (iend - ibegin .gt. 1) then
c                do j = ibegin, iend -2
c                 if (ja(j) .eq. ja(j+1)) then
c                    write(*,*) 'squack: row ', row, ' col ', ja(j)
c                    Stop
c                 end if
c                end do 
c                end if
c  end check for duplicates

c         -------------------
c         Set new row pointer
c         -------------------
          ia(row+1) = lastptr + 1

c         -------------------
c         Change last pointer
c         -------------------
          lastptr = lp

       end do
c  begin check for duplicates
c       write(*,*) 'squack: alls ok i guess'
c  end check for duplicates
       return
       end

c
c========================================================================        
c
      subroutine readf(iounit,n,rarray)
c-----------------------------------------------------------------------
c  Read in a right hand side vector f.
c-----------------------------------------------------------------------
      integer iounit, n
      real*8 rarray(n)
      read(iounit,*) (rarray(i), i = 1, n)
      return
      end
c
c=======================================================================
c
      subroutine readmt3(maxstr,job,iounit,a,colind,rwptr,b,sol,g,nrhs,
     &         work,mneed,guesol,nrow,ncol,nnz,title,key,type,ierr,
     &         chkdiag)
c-----------------------------------------------------------------------
c
c This routine reads in the transpose of a H/B file, and is a modification
c of the original SPARSKIT routine readmt.  This one differs in that the
c storage vectors a, colind, rwptr, rhs are pointers into a singly
c dimensioned array work().  At each stage, readmt3 will check the amount
c of storage provided to see if it can read the next batch of data.
c Furthermore, it can be assured that the matrix data structure includes
c space for the diagonal entries by adjusting it as we go along, if the
c argument chkdiag = .true.
c
c The pointers will be aligned on double word boundaries, if intdbl = 2.
c For this to work, however, the first word of the integer array work must
c also be aligned on a double word boundary.
c
c Note that although this is used in SPLIB to read matrices stored in
c CSR format, it works equally well on standard H/B format matrices
c (which are in CSC format).
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University, Bloomington
c     Mon Aug  8 12:25:22 EST 1994
c
c-----------------------------------------------------------------------
c
c   Control Arguments:
c   ==================
c
c maxstr =  max storage provided for a, colind, rwptr, rhs.  This will 
c           possibly limit the amount of data read in.  It should
c           be at least  n  + 1  +  nnz  integers for rwptr and colind, 
c                        nnz   real*8 for a,
c                        n     real*8's for one right hand side b,
c                        n     real*8's for one solution vector sol
c
c job	 = integer to indicate what is to be read. (note: job is an
c          input and output parameter, it can be modified on return)
c          job = 0    read the values of ncol, nrow, nnz, title, key,
c                     type and return.  Matrix is not read and arrays
c                     a, colind, rwptr, rhs are not touched.
c          job = 1    read structure only, i.e., the arrays colind and rwptr.
c          job = 2    read matrix including values, i.e., a, colind, rwptr
c          job = 3    read matrix and right hand sides: a,colind,rwptr,rhs.
c		      rhs may contain initial guesses and exact 
c                     solutions appended to the actual right hand sides.
c		      this will be indicated by the output parameter
c                     guesol [see below]. 
c                     
c iounit = logical unit number from which to read the matrix.
c work   = integer array, used for storing the vectors returned.
c next   = integer pointer, indicating next available position in
c          array work.
c
c   ============== 
c   Return Values:
c   ============== 
c
c job    = on return job may be modified to the highest job it could
c          do: if job=2 on entry but no matrix values are available it
c          is reset to job=1 on return. Similarly if job=3 but no rhs 
c          is provided then it is reset to job=2 or job=1 depending on 
c          whether or not matrix values are provided.
c          Note that no error message is triggered (i.e. ierr = 0)
c          on return in these cases.  It is therefore important to
c          compare the values of job on entry and return.
c
c a	 = (int pointer) the a matrix in the a, rwptr, colind 
c                        (column) storage format
c colind = (int pointer) column number of element a(i,j) in array a.
c rwptr  = (int pointer) pointer array. rwptr(i) points to the 
c                        beginning of column i, and rwptr(nrow+1) points to
c                        one position beyond the end of arrays a, colind.
c
c rhs    = (int pointer) real array of size nrow + 1 if available (see job)
c
c nrhs   = integer containing the number of right-hand sides found
c          each right hand side may be accompanied with an initial guess
c          and also the exact solution.
c
c guesol = a 2-character string indicating whether an initial guess 
c          (1-st character) and / or the exact solution (2-nd
c          character) is provided with the right hand side.
c	   If the first character of guesol is 'G' it means that an
c          an initial guess is provided for each right-hand side.
c          These are appended to the right hand-sides in the array rhs.
c	   If the second character of guesol is 'X' it means that an
c          exact solution is provided for each right-hand side.
c          These are  appended to the right hand-sides 
c          and the initial guesses (if any) in the array rhs.
c
c nrow   = number of rows in matrix
c ncol	 = number of columns in matrix 
c nnz	 = number of nonzero elements in A. This info is returned
c          even if there is not enough space in a, colind, rwptr, in order
c          to determine the minimum storage needed. 
c
c title  = (character*72) title of test matrix 
c key    = (character*8)  key of matrix 
c type   = (charatcer*3)  type of matrix.
c          For meaning of title, key and type refer to documentation 
c          Harwell/Boeing matrices.
c
c ierr   = integer used for error messages 
c         ierr  =  0 --> the matrix has been read normally. 
c         ierr  =  1 --> the array matrix could not be read 
c                        because inadequate storage provided.
c         ierr  =  2 --> the rhs/guess/sol could not be read.
c         ierr  =  3 --> a sparse rhs was specified.
c
c mneed  = integer; amount of extra storage needed to get all info.
c          If positive, gives amount of storage left over after reading
c          in the matrix and rhs data.  If negative, gives the additional
c          number of integer words needed to successfully read stuff in.
c 
c Notes:
c-------
c 1) The file inout must be open (and possibly rewound if necessary)
c    prior to calling readmt.
c 2) Refer to the documentation on the Harwell-Boeing formats
c    for details on the format assumed by readmt.
c    We summarize the format here for convenience.
c  
c    a) All lines in inout are assumed to be 80 character long.
c    b) The file consists of a header followed by the block of the 
c       column start pointers followed by the block of the
c       row indices, followed by the block of the real values and
c       finally the numerical values of the right-hand-side if a 
c       right hand side is supplied. 
c    c) The file starts by a header which contains four lines if no
c       right hand side is supplied and five lines otherwise.
c       * first line contains the title (72 characters long) followed by
c         the 8-character identifier (name of the matrix, called key)
c        [ A72,A8 ]
c       * second line contains the number of lines for each
c         of the following data blocks (4 of them) and the total number 
c         of lines excluding the header.
c        [5i4]
c       * the third line contains a three character string identifying
c         the type of matrices as they are referenced in the Harwell
c         Boeing documentation [e.g., rua, rsa,..] and the number of
c         rows, columns, nonzero entries.
c         [A3,11X,4I14]
c       * The fourth line contains the variable fortran format
c         for the following data blocks.
c         [2A16,2A20] 
c       * The fifth line is only present if right-hand-sides are 
c         supplied. It consists of three one character-strings containing
c         the storage format for the right-hand-sides 
c         ('F'= full,'M'=sparse=same as matrix), an initial guess 
c         indicator ('G' for yes), an exact solution indicator 
c         ('X' for yes), followed by the number of right-hand-sides
c         and then the number of row indices. 
c         [A3,11X,2I14] 
c     d) The three following blocks follow the header as described 
c        above.
c     e) In case the right hand-side are in sparse formats then 
c        the fourth block uses the same storage format as for the matrix
c        to describe the NRHS right hand sides provided, with a column
c        being replaced by a right hand side.
c-----------------------------------------------------------------------

      implicit none

c     ==================================================
c     Number of integer words per double precision word:
c     ==================================================
      include '../include/intdbl.inc'

c     Declarations of variables
c     =========================

      character title*72, key*8, type*3, ptrfmt*20, indfmt*20,
     1       valfmt*20, rhsfmt*20, rhstyp*3, guesol*2
      integer totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol,
     1     nnz, neltvl, nrhs, nrwindx
      integer mneed, maxstr, job, iounit,ierr
      integer len, nvec, n, next, remain, ndiag, row, j

c     Pointers for work array
c     -----------------------
      integer rwptr, colind , a, b, sol, g

c     Work array
c     ----------
      integer work(maxstr)

c     Insert diagonal entries or not?
c     -------------------------------
      logical chkdiag

c=======================================================================
c        Start of Executable Statements
c=======================================================================

c     -------------------
c     Read in header info
c     -------------------
      read (iounit,10) title, key, totcrd, ptrcrd, indcrd, valcrd, 
     1     rhscrd, type, nrow, ncol, nnz, neltvl, ptrfmt, indfmt, 
     2     valfmt, rhsfmt
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)

c     -----------------------------------------------
c     Read in right hand side formats if rhs provided
c     -----------------------------------------------
      if (rhscrd .gt. 0) read (iounit,11) rhstyp, nrhs , nrwindx
 11   format (a3,11x,i14,i14)


c     -----------------------
c     Anything else to read ?
c     -----------------------
      ierr = 0
      next = 1
      mneed = 0
      if (job .le. 0) return

c     ------------------------------------------------------
c     Check whether enough storage provided to read matrix.
c     Storage needed is
c        rwptr + align(rwptr) + colind + align(colind) + a,
c     where align() is a function that forces alignment on
c     double word boundaries..
c     More may be needed if diagonal entries are inserted
c     (see ndiag below).
c     ------------------------------------------------------
      remain = maxstr
      n = ncol
      mneed = n + 1      + mod(n+1,intdbl)
     &      + nnz        + mod(nnz,intdbl)
     &      + intdbl*nnz
      remain = maxstr - mneed

c     -----------------------------------------------------------
c     Note that at this point mneed only will contain the extra
c     storage needed for the matrix, exclusive of right hand side
c     data or exact solution data.
c     -----------------------------------------------------------
      if (remain .lt. 0) then
         ierr = 1
         mneed = remain
         return
      end if

c     -----------------------------------------
c     Set up pointers for rwptr, colind, and a.
c     -----------------------------------------
      rwptr = next
      colind = rwptr + n + 1

c     ------------------------------------------------------
c     Align colind so that it starts on double word boundary
c     ------------------------------------------------------
      colind = colind + mod(colind-1 ,intdbl)

c     -----------------------------------
c     Read row pointer and column numbers
c     -----------------------------------
      call rdline(iounit,ptrfmt,'int',n+1,work(rwptr),work(rwptr))
      call rdline(iounit,indfmt,'int',nnz,work(colind),work(colind))

c     ------------------------------------------------------
c     Check to see if any diagonal entries of A are missing,
c     if it is required to force diagonal entries to be
c     present.
c     ------------------------------------------------------
      ndiag = 0
      if (chkdiag) then
         do row = 1, n
            do j = work(rwptr+row-1), work(rwptr+row)-1
               if (work(colind+j-1) .eq. row) go to 20
            end do 
            ndiag = ndiag + 1
 20      continue
         end do
         nnz = ndiag + nnz
      end if 
c      Write(*,*) 'Missing diagonal entries: ', ndiag

c     ---------------------------------
c     Read values of matrix if required
c     ---------------------------------
      if (job .le. 1)  return

      if (valcrd .le. 0) then
	 job = 1
	 return
      endif

c     -------------------------
c     Now read in matrix values
c     -------------------------
      a = colind + nnz
      a = a + mod(a-1,intdbl)
      next = a

      call rdline(iounit,valfmt,'dbl',nnz-ndiag,work(a),work(a))
      next = a + intdbl*nnz

c     ------------------------
c     Read rhs if required ... 
c     ------------------------
      if (job .le. 2)  return
c     --------------------
c     ... and if available 
c     --------------------
      if ( rhscrd .le. 0) then
	 job = 2
	 return
      endif

c     ----------------------------------------------------
c     Read right-hand-side vector b, a guess if available,
c     and a true solution vector if available.
c     ----------------------------------------------------
      if (rhstyp(1:1) .eq. 'M') then 
         ierr = 3
         return
      endif
      guesol = rhstyp(2:3) 
      nvec = 1
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g')
     &              nvec=nvec+1
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x')
     &              nvec=nvec+1

c     ---------------------------------------------------
c     Check number of remaining storage words, and update
c     amount used so far:
c     ---------------------------------------------------
      remain = remain - intdbl*nvec*nrow*nrhs
      mneed  = mneed + intdbl*nvec*nrow*nrhs

      if (remain .lt. 0 ) then
         ierr = 2
         mneed = remain
         return
      endif

c     ---------------------
c     Read right-hand-sides
c     ---------------------
      len = nrhs*nrow
      b = next
      call rdline(iounit,rhsfmt,'dbl',len,work(b),work(b))
      next = b + intdbl*nrhs*nrow

c     ---------------------------------
c     Read initial guesses if available
c     ---------------------------------
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') then
         g = next
         call rdline(iounit,rhsfmt,'dbl',len,work(g),work(g))
         next = g  + intdbl*nrhs*nrow
      endif

c     ---------------------------------
c     Read exact solutions if available
c     ---------------------------------
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') then
         sol = next
         call rdline(iounit,rhsfmt,'dbl',len,work(sol),work(sol))
         next = sol  + intdbl*nrhs*nrow
      endif
c     
      return

c--------- End of readmt3 ----------------------------------------------

      end

c
c========================================================================        
c
      subroutine rdline(iounit,fmt,type,n,iarray,rarray)
      implicit none
c----------------------------------------------------------------------
c
c  Read in a line of input, specified by format fmt from IO unit
c  number iounit.  The number of words to read in is give by
c  n, and the kind (integer or double precision) by type.  One
c  of iarray or rarray will be used, depending on the type.  The
c  other can be a dummy array.
c
c----------------------------------------------------------------------
      integer iounit, n, iarray(n)
      integer i
      real*8 rarray(n)
      character*3 type
      character*20 fmt
      if (type .eq. 'int') then
         read(iounit,fmt) (iarray(i), i = 1, n)
      end if
      if (type .eq. 'dbl') then
         read(iounit,fmt) (rarray(i), i = 1, n)
      end if
      return
      end
c
c========================================================================        
c
       subroutine unpack(ireltv,  rpcprm    ,pcmeth  ,
     1                   ipcprm   ,solmeth ,krylov  ,
     2                   tol     ,maxits  , iredo,
     3                   iunit   ,
     4                   instlvl , scale, fmtd,
     5                   title, key, solver, precond)
c========================================================================        
c
c   Read in parameters that define stopping tolerances, methods to be
c   used.  Unpack information into pcmeth, ipcprm, etc.  Stop if
c   solmeth = 0.
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Wed Jan 26 13:23:46 EST 1994
c
c========================================================================        

      implicit none

      integer pcmeth, ipcprm, solmeth
      integer krylov, maxits
      integer iunit(20)
      real*8 rpcprm, tol
      logical results
      integer ireltv, iredo
      integer fmtd
      integer instlvl, scale
      character title*72, key*8
      character solver*5, precond*4

c=======================================================================
c      Start of Executable Statements
c=======================================================================

      results = (iunit(1) .ne. 0)
      read(iunit(2),*) solmeth

c     -----------------------------------------------------------
c     Sentinel value of solmeth = 0 is used for stopping the run.
c     -----------------------------------------------------------
      if (solmeth .le. 0) stop

      read(iunit(2),*) krylov
      read(iunit(2),*) pcmeth
      read(iunit(2),*) ipcprm
      read(iunit(2),*) rpcprm

      if (pcmeth .eq. 7) then
         if (rpcprm .gt. 1.0d0) ipcprm = int(rpcprm)
         if (rpcprm .lt. 1.0d0 .and. rpcprm .gt. 0.0d0) ipcprm = -2
         if (rpcprm .eq. 0.0d0) ipcprm = 0
         if (rpcprm .eq. 1.0d0) ipcprm = -1
      end if

      read(iunit(2),*) tol, ireltv
      read(iunit(2),*) maxits
      read(iunit(2),*) instlvl
      read(iunit(2),*) scale
      read(iunit(2),*) fmtd
      read(iunit(2),*) iredo

c     ------------------------------------
c     Check input parameters for validity:
c     ------------------------------------
      if (tol .lt. 0.0) then
         if (results) write(iunit(1),4040) tol
         tol = 1.0d-6
      end if
      if (rpcprm .lt. 0.0) then
         if (results) write(iunit(1),4045) rpcprm
         stop
      end if
      if (solmeth.lt. 0 .or. solmeth .gt. 14) then
         if (results) write(iunit(1),4010) solmeth
         stop
      end if
      if (pcmeth .lt. 0 .or. pcmeth .gt. 7) then
         if (results) write(iunit(1),4020) pcmeth
         stop
      end if
      if (scale .lt. 0 .or. scale .gt. 4) then
         if (results) write(iunit(1),4050) scale
         stop
      end if
c
c  Set up character strings identifying the solver/precond pair:
c
      call strings(solmeth, solver, pcmeth, precond)

c       =================
c       Format Statements
c       =================
 4010  format(' *** Invalid solution method parameter: solmeth = ',i5)
 4020  format(' *** Invalid pc method parameter: pcmeth = ',i5)
 4050  format(' *** Invalid scaling method parameter: scale = ',i2)
 4040  format(' *** Invalid tolerance: tol = ',1pe12.5,/,
     &        '     tol being reset to 1.0e-6')
 4045  format(' *** Invalid parameter: s = ',1pe12.5)

	return

c======================= End of unpack =================================

	end

c======================================================================
      subroutine prtval(iunit,type,n,iarray,rarray)
c----------------------------------------------------------------------
c
c  Print out n values of the array iarray or rarray, as
c  specified by the character array type.
c
c----------------------------------------------------------------------
      implicit none
      integer n, iarray(n), i, iunit
      real*8 rarray(n)
      character*3 type
      if (type .eq. 'int') then
         write(iunit,1000) (iarray(i), i = 1, n)
      end if
      if (type .eq. 'dbl') then
         write(iunit,1010) (rarray(i), i = 1, n)
      end if
c 1000 format(8(i7,3x))
c 1010 format(4(d18.12,3x))
 1000 format(1(i7,3x))
 1010 format(1(d24.16,3x))
      return
c----------------------------------------------------------------------
       end
c----------------------------------------------------------------------

c-----------------------------------------------------------------------
      subroutine readunf2(maxstr,nrow,ncol,nnz,a,ja,ia,iounit,
     1                  work,mneed,ierr,nrhs,rhs,title,key,type,digits)
c-----------------------------------------------------------------------
c This subroutine reads a matix stored in binary CSR format.  This differs
c from readunf in that it will read the matrix into a work vector using
c the memory manager package.
c-----------------------------------------------------------------------
c On entry:
c  maxstr: the maximum integer words allocated in work array
c  iounit: the I/O unit that opened for reading.
c  nrhs  : 0 if no right hand side vector is to be read, nonzero
c          otherwise.  nrhs is reset to zero if no rhs vector is
c          available to be read.  This is not an error condition, so
c          if you request a rhs to be read, be sure to check nrhs on
c          return to see if one was available.
c  digits: number of significant digits for representing floats.
c On return:
c    nrow: the actual number of rows in array.
c    ncol: the actual number of cols in array.
c     nnz: the actual number of non-zero entries.
c    nrhs: the actual number of right hand side vectors read (0 or 1)
c ia,ja,a: pointers in work array to where corr arrays go.
c    ierr: a error code, it's same as that used in reaadsk
c          0 -- OK
c          1 -- error in reading iounit
c          2 -- end-of-file reached while reading data file
c          3 -- not enough storage provided; check mneed
c On error:
c     mneed contains additional storage needed if not enough was provided.
c-----------------------------------------------------------------------
c
      integer intdbl
      parameter (intdbl = 2)
      integer nrow,ncol, iounit, nnz, k, digits
      integer work(maxstr), maxstr, remain
      integer ia, ja, a, rhs, next, mneed
      character title*72, key*8, type*3
c
      ierr = 0
      rewind iounit
c      
      read(unit=iounit, err=1000, end=1010) nrow, ncol
      read(unit=iounit, err=1000, end=1010) title, key, type,
     &                                       digits, nrhs

      mneed = 0
      next = 1
      ia = next
      if (nrow+1 .gt. maxstr) then
         ierr = 3
         mneed = maxstr - nrow-1
         return
      end if 
      read(unit=iounit, err=1000, end=1010) (work(ia+k),k=0,nrow)
c
      nnz = work(ia+nrow) - 1
      next = next + nrow + 1
      next = next + mod(next-1, intdbl)
      ja = next
      remain = maxstr - next
      if (remain .lt. (intdbl+1)*nnz) then
         ierr = 3
         mneed = remain - (intdbl+1)*nnz
         return
      end if 
      read(unit=iounit, err=1000, end=1010) (work(ja+k),k=0,nnz-1)

      next = next + nnz
      next = next + mod(next-1, intdbl)
      a = next
      remain = maxstr - next

      call rlunf(work(a), nnz, iounit, ierr)
      if (ierr .ne. 0 ) return
      if (nrhs .le. 0)  return

      next = next + intdbl*nnz
      rhs = next
      remain = maxstr - next
      if (remain .lt. intdbl*nnz) then
         ierr = 3
         mneed = remain - intdbl*nnz
         return
      end if 
      call rlunf(work(rhs), nrow, iounit, ierr)
c
      return
c
c     error handling
c
 1000 ierr = 1
      return
 1010 ierr = 2
      return
      end
c---------end of readunf2 ----------------------------------------------
      subroutine rlunf(a, m, iounit, ierr)
c
c  Read in a double precision array from iounit, in binary format.
c  ierr = 0 if all's ok
c  ierr = 1 if error found in reading
c  ierr = 2 if end reached before m values read.
c
      integer m, iounit, ierr
      real*8 a(*)
      ierr  = 0
      read(unit=iounit, err=1000, end=1010) (a(k),k=1,m)
      return
 1000 ierr = 1
      return
 1010 ierr = 2
      return
      end
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
      subroutine prtmt (nrow,ncol,a,ja,ia,rhs,guesol,title,key,type,
     1     ifmt,job,iounit)
c-----------------------------------------------------------------------
c writes a matrix in Harwell-Boeing format into a file.
c assumes that the matrix is stored in COMPRESSED SPARSE COLUMN FORMAT.
c some limited functionality for right hand sides. 
c Author: Youcef Saad - Date: Sept., 1989 - updated Oct. 31, 1989 to
c cope with new format. 
c-----------------------------------------------------------------------
c on entry:
c---------
c nrow   = number of rows in matrix
c ncol	 = number of columns in matrix 
c a	 = real*8 array containing the values of the matrix stored 
c          columnwise
c ja 	 = integer array of the same length as a containing the column
c          indices of the corresponding matrix elements of array a.
c ia     = integer array of containing the pointers to the beginning of 
c	   the row in arrays a and ja.
c rhs    = real array  containing the right-hand-side (s) and optionally
c          the associated initial guesses and/or exact solutions
c          in this order. See also guesol for details. the vector rhs will
c          be used only if job .gt. 2 (see below). Only full storage for
c          the right hand sides is supported. 
c
c guesol = a 2-character string indicating whether an initial guess 
c          (1-st character) and / or the exact solution (2-nd)
c          character) is provided with the right hand side.
c	   if the first character of guesol is 'G' it means that an
c          an intial guess is provided for each right-hand sides. 
c          These are assumed to be appended to the right hand-sides in 
c          the array rhs.
c	   if the second character of guesol is 'X' it means that an
c          exact solution is provided for each right-hand side.
c          These are assumed to be appended to the right hand-sides 
c          and the initial guesses (if any) in the array rhs.
c
c title  = character*72 = title of matrix test ( character a*72 ).
c key    = character*8  = key of matrix 
c type   = charatcer*3  = type of matrix.
c
c ifmt	 = integer specifying the format chosen for the real values
c	   to be output (i.e., for a, and for rhs-guess-sol if 
c          applicable). The meaning of ifmt is as follows.
c	  * if (ifmt .lt. 100) then the D descriptor is used,
c           format Dd.m, in which the length (m) of the mantissa is 
c           precisely the integer ifmt (and d = ifmt+6)
c	  * if (ifmt .gt. 100) then prtmt will use the 
c           F- descriptor (format Fd.m) in which the length of the 
c           mantissa (m) is the integer mod(ifmt,100) and the length 
c           of the integer part is k=ifmt/100 (and d = k+m+2)
c	    Thus  ifmt= 4   means  D10.4  +.xxxxD+ee    while
c	          ifmt=104  means  F7.4   +x.xxxx
c	          ifmt=205  means  F9.5   +xx.xxxxx
c	    Note: formats for ja, and ia are internally computed.
c
c job	 = integer to indicate whether matrix values and
c	   a right-hand-side is available to be written
c          job = 1   write srtucture only, i.e., the arrays ja and ia.
c          job = 2   write matrix including values, i.e., a, ja, ia
c          job = 3   write matrix and one right hand side: a,ja,ia,rhs.
c	   job = nrhs+2 write matrix and nrhs successive right hand sides
c	   Note that there cannot be any right-hand-side if the matrix
c	   has no values. Also the initial guess and exact solutions when 
c          provided are for each right hand side. For example if nrhs=2 
c          and guesol='GX' there are 6 vectors to write.
c          
c
c iounit = logical unit number where to write the matrix into.
c
c on return:
c---------- 
c the matrix a, ja, ia will be written in output unit iounit
c in the Harwell-Boeing format. None of the inputs is modofied.
c  
c Notes: 1) This code attempts to pack as many elements as possible per
c        80-character line. 
c        2) this code attempts to avoid as much as possible to put
c        blanks in the formats that are written in the 4-line header
c	 (This is done for purely esthetical reasons since blanks
c        are ignored in format descriptors.)
c        3) sparse formats for right hand sides and guesses are not
c        supported.
c-----------------------------------------------------------------------
      character title*72,key*8,type*3,ptrfmt*16,indfmt*16,valfmt*20,
     *	        guesol*2, rhstyp*3
      integer totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol,
     1     nnz, nrhs, len, nperli
      integer ja(*), ia(*) 	
      real*8 a(*),rhs(*)
c--------------
c     compute pointer format
c--------------
      nnz    = ia(ncol+1) -1
      len    = int ( alog10(0.1+real(nnz+1))) + 1
      nperli = 80/len
      ptrcrd = ncol/nperli + 1
      if (len .gt. 9) then
         assign 101 to ix
      else
         assign 100 to ix
      endif
      write (ptrfmt,ix) nperli,len
 100  format(1h(,i2,1Hi,i1,1h) )
 101  format(1h(,i2,1Hi,i2,1h) )
c----------------------------
c compute ROW index format
c----------------------------
      len    = int ( alog10(0.1+real(nrow) )) + 1
      nperli = min0(80/len,nnz)
      indcrd = (nnz-1)/nperli+1
      write (indfmt,100) nperli,len
c---------------
c compute values and rhs format (using the same for both)
c--------------- 
      valcrd	= 0
      rhscrd  = 0
c quit this part if no values provided.
      if (job .le. 1) goto 20
c     
      if (ifmt .ge. 100) then
         ihead = ifmt/100
         ifmt = ifmt-100*ihead
         len = ihead+ifmt+2
         nperli = 80/len
c     
         if (len .le. 9 ) then
            assign 102 to ix
         elseif (ifmt .le. 9) then
            assign 103 to ix
         else 
            assign 104 to ix
         endif
c     
         write(valfmt,ix) nperli,len,ifmt
 102     format(1h(,i2,1hf,i1,1h.,i1,1h) )
 103     format(1h(,i2,1hf,i2,1h.,i1,1h) )
 104     format(1h(,i2,1hf,i2,1h.,i2,1h) )
C
      else
         len = ifmt + 6
         nperli = 80/len
c     try to minimize the blanks in the format strings.
         if (nperli .le. 9) then
	    if (len .le. 9 ) then
	       assign 105 to ix
	    elseif (ifmt .le. 9) then
	       assign 106 to ix
	    else 
	       assign 107 to ix
	    endif
	 else 
	    if (len .le. 9 ) then
	       assign 108 to ix
	    elseif (ifmt .le. 9) then
	       assign 109 to ix
	    else 
               assign 110 to ix
            endif
         endif
c-----------
         write(valfmt,ix) nperli,len,ifmt
 105     format(1h(,i1,1hd,i1,1h.,i1,1h) )
 106     format(1h(,i1,1hd,i2,1h.,i1,1h) )
 107     format(1h(,i1,1hd,i2,1h.,i2,1h) )
 108     format(1h(,i2,1hd,i1,1h.,i1,1h) )
 109     format(1h(,i2,1hd,i2,1h.,i1,1h) )
 110     format(1h(,i2,1hd,i2,1h.,i2,1h) )
c     
      endif 	    
      valcrd = (nnz-1)/nperli+1
      nrhs   = job -2
      if (nrhs .ge. 1) then
         i = (nrhs*nrow-1)/nperli+1
         rhscrd = i
         if (guesol(1:1) .eq. 'g') guesol(1:1) = 'G'
         if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g')
     +       rhscrd = rhscrd+i
         if (guesol(2:2) .eq. 'x') guesol(2:2) = 'X'
         if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x')
     +       rhscrd = rhscrd+i
         rhstyp = 'F'//guesol
      endif 
 20   continue
c     
      totcrd = ptrcrd+indcrd+valcrd+rhscrd
c     write 4-line or five line header
      write(iounit,10) title,key,totcrd,ptrcrd,indcrd,valcrd,
     1     rhscrd,type,nrow,ncol,nnz,nrhs,ptrfmt,indfmt,valfmt,valfmt
c-----------------------------------------------------------------------
      if (nrhs .ge. 1) write (iounit,11) rhstyp, nrhs
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)
 11   format(A3,11x,i4)
c     
      write(iounit,ptrfmt) (ia (i), i = 1, ncol+1)
      write(iounit,indfmt) (ja (i), i = 1, nnz)
      if (job .le. 1) return
      write(iounit,valfmt) (a(i), i = 1, nnz)
      if (job .le. 2) return 
      len = nrow*nrhs 
      next = 1
      iend = len
      write(iounit,valfmt) (rhs(i), i = next, iend)
c     
c     write initial guesses if available
c     
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') then
         next = next+len
         iend = iend+ len
         write(iounit,valfmt) (rhs(i), i = next, iend)
      endif
c     
c     write exact solutions if available
c     
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x')then
         next = next+len
         iend = iend+ len
         write(iounit,valfmt) (rhs(i), i = next, iend)
      endif
c     
      return
c----------end of prtmt ------------------------------------------------ 
c-----------------------------------------------------------------------
      end
